@@ -64,7 +64,9 @@
 
 /*#define DEBUG_BY_READING_PAGES*/
 
-#define CHUNKSIZE 8500
+#define CHUNKSIZE 65536
+
+#define OGGZ_READ_EMPTY (-404)
 
 #define oggz_off_t long
 
@@ -80,6 +82,9 @@ oggz_read_init (OGGZ * oggz)
   reader->read_packet = NULL;
   reader->read_user_data = NULL;
 
+  reader->read_page = NULL;
+  reader->read_page_user_data = NULL;
+
   reader->current_unit = 0;
 
   return oggz;
@@ -90,6 +95,7 @@ oggz_read_close (OGGZ * oggz)
 {
   OggzReader * reader = &oggz->x.reader;
 
+  ogg_stream_clear (&reader->ogg_stream);
   ogg_sync_clear (&reader->ogg_sync);
 
   return oggz;
@@ -129,6 +135,40 @@ oggz_set_read_callback (OGGZ * oggz, long serialno,
   return 0;
 }
 
+int
+oggz_set_read_page (OGGZ * oggz, long serialno, OggzReadPage read_page,
+		    void * user_data)
+{
+  OggzReader * reader;
+  oggz_stream_t * stream;
+
+  if (oggz == NULL) return OGGZ_ERR_BAD_OGGZ;
+
+  reader =  &oggz->x.reader;
+
+  if (oggz->flags & OGGZ_WRITE) {
+    return OGGZ_ERR_INVALID;
+  }
+
+  if (serialno == -1) {
+    reader->read_page = read_page;
+    reader->read_page_user_data = user_data;
+  } else {
+    stream = oggz_get_stream (oggz, serialno);
+#if 0
+    if (stream == NULL) return OGGZ_ERR_BAD_SERIALNO;
+#else
+    if (stream == NULL)
+      stream = oggz_add_stream (oggz, serialno);
+#endif
+
+    stream->read_page = read_page;
+    stream->read_page_user_data = user_data;
+  }
+
+  return 0;
+}
+
 /*
  * oggz_get_next_page_7 (oggz, og, do_read)
  *
@@ -233,7 +273,7 @@ oggz_read_sync (OGGZ * oggz)
       serialno = reader->current_serialno;
 
       stream = oggz_get_stream (oggz, serialno);
-      
+
       if (stream == NULL) {
 	/* new stream ... check bos etc. */
 	if ((stream = oggz_add_stream (oggz, serialno)) == NULL) {
@@ -258,10 +298,12 @@ oggz_read_sync (OGGZ * oggz)
 	  oggz_auto (oggz, op, serialno, NULL);
 	}
 
+#if 0 /* set unit on last packet of page */
 	if ((oggz->metric || stream->metric) && granulepos != -1) {
 	  reader->current_unit = oggz_get_unit (oggz, serialno, granulepos);
 	}
-#ifndef DEBUG_BY_READING_PAGES
+#endif
+
 	if (stream->read_packet) {
 	  cb_ret =
 	    stream->read_packet (oggz, op, serialno, stream->read_user_data);
@@ -269,7 +311,6 @@ oggz_read_sync (OGGZ * oggz)
 	  cb_ret =
 	    reader->read_packet (oggz, op, serialno, reader->read_user_data);
 	}
-#endif /* DEBUG_BY_READING_PAGES */
       }
       else
 	break;
@@ -277,16 +318,16 @@ oggz_read_sync (OGGZ * oggz)
     }
 
     /* If we've got a stop already, don't read more data in */
-    if (cb_ret != 0) return cb_ret;
+    if (cb_ret == OGGZ_STOP_OK || cb_ret == OGGZ_STOP_ERR) return cb_ret;
 
     if(oggz_get_next_page_7 (oggz, &og) < 0)
-      return -404; /* eof. leave unitialized */
+      return OGGZ_READ_EMPTY; /* eof. leave uninitialized */
 
     serialno = ogg_page_serialno (&og);
     reader->current_serialno = serialno; /* XXX: maybe not necessary */
 
     stream = oggz_get_stream (oggz, serialno);
-      
+
     if (stream == NULL) {
       /* new stream ... check bos etc. */
       if ((stream = oggz_add_stream (oggz, serialno)) == NULL) {
@@ -296,25 +337,24 @@ oggz_read_sync (OGGZ * oggz)
     }
     os = &stream->ogg_stream;
 
-#ifdef DEBUG_BY_READING_PAGES
     {
-      ogg_packet op_debug;
-      op_debug.packet = og.body;
-      op_debug.bytes = og.body_len;
-      op_debug.b_o_s = ogg_page_bos (&og);
-      op_debug.e_o_s = ogg_page_eos (&og);
-      op_debug.granulepos = ogg_page_granulepos (&og);
-      op_debug.packetno = ogg_page_packets (&og);
+      ogg_int64_t granulepos;
 
-      if (stream->read_packet) {
-	cb_ret = stream->read_packet (oggz, &op_debug, serialno,
-				      stream->read_user_data);
-      } else if (reader->read_packet) {
-	cb_ret = reader->read_packet (oggz, &op_debug, serialno,
-				      reader->read_user_data);
+      granulepos = ogg_page_granulepos (&og);
+      if ((oggz->metric || stream->metric) && granulepos != -1) {
+	reader->current_unit = oggz_get_unit (oggz, serialno, granulepos);
+      } else if (granulepos == 0) {
+	reader->current_unit = 0;
       }
     }
-#endif
+
+    if (stream->read_page) {
+      cb_ret =
+	stream->read_page (oggz, &og, serialno, stream->read_page_user_data);
+    } else if (reader->read_page) {
+      cb_ret = reader->read_page (oggz, &og, serialno,
+				  reader->read_page_user_data);
+    }
 
 #if 0
     /* bitrate tracking; add the header's bytes here, the body bytes
@@ -346,11 +386,23 @@ oggz_read (OGGZ * oggz, long n)
 
   cb_ret = oggz_read_sync (oggz);
 
-  /* If there's nothing to read yet, don't flag an error */
-  if (reader->current_unit == 0 && cb_ret == -404) cb_ret = 0;
+#if 0
+  if (cb_ret == OGGZ_READ_EMPTY) {
+    /* If there's nothing to read yet, don't return 0 (eof) */
+    if (reader->current_unit == 0) cb_ret = 0;
+    else {
+#if 0
+      printf ("oggz_read: EMPTY, current_unit %ld != 0\n",
+	      reader->current_unit);
+      return 0;
+#endif
+    }
+  }
+#endif
 
-  while (cb_ret != -1 && cb_ret != 1 && bytes_read > 0 && remaining > 0) {
-    bytes = MIN (remaining, 4096);
+  while (cb_ret != OGGZ_STOP_ERR && cb_ret != OGGZ_STOP_OK &&
+	 bytes_read > 0 && remaining > 0) {
+    bytes = MIN (remaining, CHUNKSIZE);
     buffer = ogg_sync_buffer (&reader->ogg_sync, bytes);
     if ((bytes_read = (long) oggz_io_read (oggz, buffer, bytes)) == 0) {
       /* schyeah! */
@@ -359,15 +411,37 @@ oggz_read (OGGZ * oggz, long n)
       return OGGZ_ERR_SYSTEM;
     }
 
-    ogg_sync_wrote (&reader->ogg_sync, bytes_read);
-
-    remaining -= bytes_read;
-    nread += bytes_read;
-
-    cb_ret = oggz_read_sync (oggz);
+    if (bytes_read > 0) {
+      ogg_sync_wrote (&reader->ogg_sync, bytes_read);
+      
+      remaining -= bytes_read;
+      nread += bytes_read;
+      
+      cb_ret = oggz_read_sync (oggz);
+    }
   }
 
-  if (cb_ret == -1) oggz_purge (oggz);
+  if (cb_ret == OGGZ_STOP_ERR) oggz_purge (oggz);
+
+  /* Don't return 0 unless it's actually an EOF condition */
+  if (nread == 0) {
+    switch (bytes_read) {
+    case OGGZ_ERR_IO_AGAIN:
+    case OGGZ_ERR_SYSTEM:
+      return bytes_read; break;
+    default: break;
+    }
+
+    switch (cb_ret) {
+    case OGGZ_CONTINUE: case OGGZ_READ_EMPTY: 
+#ifdef DEBUG
+      printf ("oggz_read: nread==0, cb_ret==%d, returning 0\n", cb_ret);
+#endif
+      return 0; break;
+    case OGGZ_STOP_ERR: return OGGZ_ERR_READ_STOP_ERR; break;
+    case OGGZ_STOP_OK: default: return OGGZ_ERR_READ_STOP_OK; break;
+    }
+  }
 
   return nread;
 }
@@ -391,8 +465,13 @@ oggz_read_input (OGGZ * oggz, unsigned char * buf, long n)
 
   cb_ret = oggz_read_sync (oggz);
 
-  /* If there's nothing to read yet, don't flag an error */
-  if (reader->current_unit == 0 && cb_ret == -404) cb_ret = 0;
+#if 0
+  if (cb_ret == OGGZ_READ_EMPTY) {
+    /* If there's nothing to read yet, don't return 0 (eof) */
+    if (reader->current_unit == 0) cb_ret = 0;
+    else return 0;
+  }
+#endif
 
   while (cb_ret != -1 && cb_ret != 1 && /* !oggz->eos && */ remaining > 0) {
     bytes = MIN (remaining, 4096);
@@ -404,10 +483,25 @@ oggz_read_input (OGGZ * oggz, unsigned char * buf, long n)
     remaining -= bytes;
     nread += bytes;
 
-    cb_ret = oggz_read_sync (oggz);    
+    cb_ret = oggz_read_sync (oggz);
   }
 
-  if (cb_ret == -1) oggz_purge (oggz);
+  if (cb_ret == OGGZ_STOP_ERR) oggz_purge (oggz);
+
+  /* Don't return 0 unless it's actually an EOF condition */
+  if (nread == 0) {
+    switch (cb_ret) {
+    case OGGZ_CONTINUE: return 0; break;
+    case OGGZ_READ_EMPTY:
+#ifdef DEBUG
+      printf ("oggz_read_input: OUT EMPTY\n");
+#endif
+      return OGGZ_ERR_READ_STOP_OK;
+      break;
+    case OGGZ_STOP_ERR: return OGGZ_ERR_READ_STOP_ERR; break;
+    case OGGZ_STOP_OK: default: return OGGZ_ERR_READ_STOP_OK; break;
+    }
+  }
 
   return nread;
 }
@@ -560,8 +654,14 @@ oggz_get_next_page (OGGZ * oggz, ogg_page * og)
       }
 
       if (bytes == 0) {
+#ifdef DEBUG
+	printf ("get_next_page: bytes == 0, returning -2\n");
+#endif
 	return -2;
       } else if (oggz->file && feof (oggz->file)) {
+#ifdef DEBUG
+	printf ("get_next_page: feof (oggz->file), returning -2\n");
+#endif
 	clearerr (oggz->file);
 	return -2;
       }
@@ -601,18 +701,21 @@ oggz_get_next_start_page (OGGZ * oggz, ogg_page * og)
   oggz_off_t page_offset;
   int found = 0;
 
-  while (!found) {
+  do {
     page_offset = oggz_get_next_page (oggz, og);
 
     /* Return this value if one of the following conditions is met:
      *
      *   page_offset < 0     : error or EOF
      *   page_offset == 0    : start of stream
-     *   !ogg_page_continued : start of page
+     *   !ogg_page_continued : page contains start of a packet
+     *   ogg_page_packets > 1: page contains start of a packet
      */
-    if (page_offset <= 0 || !ogg_page_continued (og))
+    if (page_offset <= 0 || !ogg_page_continued (og) ||
+	ogg_page_packets (og) > 1)
       found = 1;
   }
+  while (!found);
 
   return page_offset;
 }
@@ -623,6 +726,7 @@ oggz_get_prev_start_page (OGGZ * oggz, ogg_page * og,
 {
   oggz_off_t offset_at, offset_start;
   oggz_off_t page_offset, prev_offset = 0;
+  ogg_int64_t unit_at;
   long granule_at = -1;
 
 #if 0
@@ -644,10 +748,10 @@ oggz_get_prev_start_page (OGGZ * oggz, ogg_page * og,
 
 #ifdef DEBUG
 
-    printf ("[A] offset_at: @%ld\toffset_start: @%ld\n",
+    printf ("get_prev_start_page: [A] offset_at: @%ld\toffset_start: @%ld\n",
 	    offset_at, offset_start);
 
-    printf ("*** get_prev_start_page: seeked to %ld\n", offset_start);
+    printf ("get_prev_start_page: seeked to %ld\n", offset_start);
 #endif
 
     page_offset = 0;
@@ -657,13 +761,18 @@ oggz_get_prev_start_page (OGGZ * oggz, ogg_page * og,
 
       page_offset = oggz_get_next_start_page (oggz, og);
       if (page_offset == -1) return -1;
-      if (page_offset == -2) break;
+      if (page_offset == -2) {
+#ifdef DEBUG
+	printf ("*** get_prev_start_page: page_offset = -2\n");
+#endif
+	break;
+      }
 
       granule_at = (long)ogg_page_granulepos (og);
 
 #ifdef DEBUG
-      printf ("\tGOT page (%ld) @%ld\tat @%ld\n", granule_at,
-	      page_offset, offset_at);
+      printf ("get_prev_start_page: GOT page (%ld) @%ld\tat @%ld\n",
+	      granule_at, page_offset, offset_at);
 #endif
 
       /* Need to stash the granule and serialno of this page because og
@@ -677,7 +786,7 @@ oggz_get_prev_start_page (OGGZ * oggz, ogg_page * og,
     } while (page_offset >= 0 && page_offset < offset_at);
 
 #ifdef DEBUG
-    printf ("[B] offset_at: @%ld\toffset_start: @%ld\n"
+    printf ("get_prev_start_page: [B] offset_at: @%ld\toffset_start: @%ld\n"
 	    "prev_offset: @%ld\tpage_offset: @%ld\n",
 	    offset_at, offset_start, prev_offset, page_offset);
 #endif
@@ -686,6 +795,10 @@ oggz_get_prev_start_page (OGGZ * oggz, ogg_page * og,
 
   } while (offset_at > 0 && prev_offset == 0);
 
+  unit_at = oggz_get_unit (oggz, *serialno, *granule);
+  offset_at = oggz_reset (oggz, prev_offset, unit_at, SEEK_SET);
+  if (offset_at == -1) return -1;
+
   if (offset_at > 0)
     return prev_offset;
   else
@@ -706,7 +819,7 @@ oggz_scan_for_page (OGGZ * oggz, ogg_page * og, ogg_int64_t unit_target,
   printf (" SCANNING from %ld...", offset_begin);
 #endif
 
-  while (1) {
+  for ( ; ; ) {
     offset_at = oggz_seek_raw (oggz, offset_begin, SEEK_SET);
     if (offset_at == -1) return -1;
 
@@ -790,7 +903,9 @@ oggz_scan_for_page (OGGZ * oggz, ogg_page * og, ogg_int64_t unit_target,
   return offset_at;
 }
 
-static long
+#define GUESS_MULTIPLIER (1<<16)
+
+static ogg_int64_t
 oggz_seek_set (OGGZ * oggz, ogg_int64_t unit_target)
 {
   OggzReader * reader = &oggz->x.reader;
@@ -801,7 +916,7 @@ oggz_seek_set (OGGZ * oggz, ogg_int64_t unit_target)
   ogg_int64_t granule_at;
   ogg_int64_t unit_at, unit_begin = 0, unit_end = -1;
   long serialno;
-  double guess_ratio;
+  ogg_int64_t guess_ratio;
   ogg_page * og;
 
   if (oggz == NULL) {
@@ -813,7 +928,7 @@ oggz_seek_set (OGGZ * oggz, ogg_int64_t unit_target)
     return -1;
   }
 
-  if (oggz->file != NULL) { 
+  if (oggz->file != NULL) {
     if ((fd = fileno (oggz->file)) == -1) {
       /*oggz_set_error (oggz, OGGZ_ERR_SYSTEM);*/
       return -1;
@@ -878,7 +993,7 @@ oggz_seek_set (OGGZ * oggz, ogg_int64_t unit_target)
 
   og = &oggz->current_page;
 
-  while (1) {
+  for ( ; ; ) {
 
 #ifdef DEBUG
     printf ("oggz_read_seek (%ld): (%ld - %ld) [%ld - %ld]\t",
@@ -896,16 +1011,17 @@ oggz_seek_set (OGGZ * oggz, ogg_int64_t unit_target)
 	printf ("*G2*");
 #endif
 	guess_ratio =
-	  (double)(unit_target - unit_begin) /
-	  (double)(unit_at - unit_begin);
+	  GUESS_MULTIPLIER * (unit_target - unit_begin) /
+	  (unit_at - unit_begin);
 
 #ifdef DEBUG
-	printf ("\nguess_ration %f = (%ld - %ld) / (%ld - %ld)\n",
+	printf ("\nseek_set: guess_ratio %lld = (%lld - %lld) / (%lld - %lld)\n",
 		guess_ratio, unit_target, unit_begin, unit_at, unit_begin);
 #endif
 
 	offset_guess = offset_begin +
-	  (oggz_off_t)((offset_at - offset_begin) * guess_ratio);
+	  (oggz_off_t)(((offset_at - offset_begin) * guess_ratio) /
+		       GUESS_MULTIPLIER);
       }
     } else if (unit_end <= unit_begin) {
 #ifdef DEBUG
@@ -915,11 +1031,12 @@ oggz_seek_set (OGGZ * oggz, ogg_int64_t unit_target)
     } else {
 #if 1
       guess_ratio =
-	(double)(unit_target - unit_begin) /
-	(double)(unit_end - unit_begin);
+	GUESS_MULTIPLIER * (unit_target - unit_begin) /
+	(unit_end - unit_begin);
 
       offset_guess = offset_begin +
-	(oggz_off_t)((offset_end - offset_begin) * guess_ratio);
+	(oggz_off_t)(((offset_end - offset_begin) * guess_ratio) /
+		     GUESS_MULTIPLIER);
 
       /*
       if (offset_guess <= offset_begin) {
@@ -946,6 +1063,11 @@ oggz_seek_set (OGGZ * oggz, ogg_int64_t unit_target)
     printf ("\n");
 #endif
 
+    if (unit_end == -1 && offset_next == -2) { /* reached eof, backtrack */
+      offset_next = oggz_get_prev_start_page (oggz, og, &granule_at,
+					      &serialno);
+    }
+
     if (offset_next < 0) {
       goto notfound;
     }
@@ -1009,7 +1131,7 @@ oggz_seek_set (OGGZ * oggz, ogg_int64_t unit_target)
   return -1;
 }
 
-static long
+static ogg_int64_t
 oggz_seek_end (OGGZ * oggz, ogg_int64_t unit_offset)
 {
   oggz_off_t offset_orig, offset_at, offset_end;
@@ -1033,10 +1155,10 @@ oggz_seek_end (OGGZ * oggz, ogg_int64_t unit_offset)
     oggz_reset (oggz, offset_orig, -1, SEEK_SET);
     return -1;
   }
-  
+
 #ifdef DEBUG
-  printf ("*** oggz_seek_end: found packet (%ld) at @%ld\n",
-	  unit_end, offset_end);
+  printf ("*** oggz_seek_end: found packet (%ld) at @%ld [%ld]\n",
+	  unit_end, offset_end, granulepos);
 #endif
 
   return oggz_seek_set (oggz, unit_end + unit_offset);
@@ -1052,13 +1174,13 @@ oggz_seek (OGGZ * oggz, oggz_off_t offset, int whence)
   if (oggz->flags & OGGZ_WRITE) {
     return -1;
   }
-  
+
   if (offset == 0) units = 0;
 
   return (off_t)oggz_reset (oggz, offset, units, whence);
 }
 
-long
+ogg_int64_t
 oggz_seek_units (OGGZ * oggz, ogg_int64_t units, int whence)
 {
   OggzReader * reader = &oggz->x.reader;
@@ -1077,7 +1199,7 @@ oggz_seek_units (OGGZ * oggz, ogg_int64_t units, int whence)
   case SEEK_SET:
     return oggz_seek_set (oggz, units);
     break;
-  case SEEK_CUR: 
+  case SEEK_CUR:
     units += reader->current_unit;
     return oggz_seek_set (oggz, units);
     break;
@@ -1108,6 +1230,18 @@ oggz_seek_packets (OGGZ * oggz, long serialno, long packets, int whence)
 #include <ogg/ogg.h>
 #include "oggz_private.h"
 
+OGGZ *
+oggz_read_init (OGGZ * oggz)
+{
+  return NULL;
+}
+
+OGGZ *
+oggz_read_close (OGGZ * oggz)
+{
+  return NULL;
+}
+
 int
 oggz_set_read_callback (OGGZ * oggz, long serialno,
 			OggzReadPacket read_packet, void * user_data)
